package dataSetDescribers;

import java.io.File;
import java.util.ArrayList;
import java.util.List;
import utils.ConfigReader;

public class PincProbe extends ProbeAbstract
{		
	private AbstractSpikedControls spikeDetails = null;
	private final List<File> baselineCelFiles;
	private final List<File> experimentCelFiles;

	@Override
	public final String getExperimentDescripton()
    {
		StringBuffer buff = new StringBuffer( getShortExperimentName() + "_");
		
		if( isUseStatisticLevelNormalization() )
			buff.append( "PINC" );
		
		return buff.toString();
    }
	
    @Override
    public File getCdfFile() throws Exception
    {
    	File cdfFile = new File( ConfigReader.getCDFFile() );
    	
    	if( ! cdfFile.exists())
    		throw new Exception("Could not find cdf file " + cdfFile.getAbsolutePath());
    	
    	return cdfFile;
    }
    

    @Override
    public String getChipId() throws Exception
    {
    	File cdfFile = getCdfFile();
    	
    	String returnVal = cdfFile.getName();
    
    	if( returnVal.toUpperCase().endsWith(".CDF"))
    		returnVal = returnVal.substring(0, returnVal.length()-4);
    	
        return ConfigReader.getCDFFile();
    }
	    
		
	public PincProbe( List<File> baselineCelFiles, 
			List<File> experimentCelFiles ) 
	{
		this.baselineCelFiles = baselineCelFiles;
		this.experimentCelFiles = experimentCelFiles;
	}
	
	private static File getCELFile(String name, boolean isBaseline ) throws Exception
	{
		if (isBaseline)
		{
			return new File(
					ConfigReader.getBaselineCelsDirectory()
					+ File.separator + name );
		}
		else
		{
			return new File(
					ConfigReader.getExperimentCelsDirectory()
					+ File.separator + name );
		}
	}
	
	
	@Override
	public List<File> getBaselineCelFiles()
	{
		return baselineCelFiles;
	}

	@Override
	public List<File> getExperimentCelFiles()
	{
		return experimentCelFiles;
	}

	@Override
	int getNumExpectedProbes() throws Exception
	{
		return ConfigReader.getExpectedNumberOfProbe();
	}
	
	private String getShortName(File file)
	{
		String name = file.getName();
		
		if( name.toUpperCase().	endsWith(".CEL"))
			name = name.substring(0, name.length() - 4);
		
		return name;
	}
	
	public static void main(String[] args) throws Exception
	{
		List<PincProbe> list = getAllPossibleComparions();
		
		for( PincProbe lsg : list )
		{
			System.out.println(lsg.getShortExperimentName());
			
			for( File f : lsg.getBaselineCelFiles() )
				if( ! f.exists())
					throw new Exception(f.getAbsolutePath());
			

			for( File f : lsg.getExperimentCelFiles() )
				if( ! f.exists())
					throw new Exception(f.getAbsolutePath());
		}
			
	}
	
	
	@Override
	public String getShortExperimentName()
	{
		StringBuffer buff = new StringBuffer();
		
		for( File file :  getExperimentCelFiles() )
			buff.append( getShortName(file));
		
		buff.append("_VS_");
		
		for( File file :  getBaselineCelFiles() )
			buff.append( getShortName(file));
		
		return buff.toString();
	}
	
	 /**
     *   Lazy initialized and not thread safe
     */
    public AbstractSpikedControls getSpikeDetails()
    {
        if ( spikeDetails == null)
        {
            spikeDetails = new HGU133ALatinSquare();
        }
        return spikeDetails;
    }
    
    private static File getDirOrThrow(String filePath) throws Exception
    {
    	File aDir = getFileOrThrow(filePath);
    	
    	if( ! aDir.isDirectory())
    		throw new Exception(aDir.getAbsolutePath() + " is not a directory");
    	
    	return aDir;
    }
    
    private static File getFileOrThrow(String filePath) throws Exception
    {
    	File aFile = new File(filePath);
    	
    	if( ! aFile.exists())
    		throw new Exception( "Could not find " + aFile.getAbsolutePath());
    	
    	return aFile;
    }
    
    
    public static List<PincProbe> getAllPossibleComparions() throws Exception
	{
		// Directories where cells are
    	File controlDir = getDirOrThrow(ConfigReader.getBaselineCelsDirectory());
		File exptDir = getDirOrThrow(ConfigReader.getExperimentCelsDirectory());
    	
		//our return list
		List<PincProbe> returnList = new ArrayList<PincProbe>();
		
		List<String> setcontrol = new ArrayList<String>();
		List<String> setExpt = new ArrayList<String>();
		
		//fill baseline cel list
		for (String fileName: controlDir.list())
			if( fileName.toUpperCase().endsWith(".CEL"))
			{
				setcontrol.add(fileName);
			}
		
		if( setcontrol.size() == 0)
			throw new Exception("Could not find any cel files in baseline directory");
			
		// fill expt cel list
		for (String fileName: exptDir.list())
			if( fileName.toUpperCase().endsWith(".CEL"))
			{
				setExpt.add(fileName);
			}
		
		if( setExpt.size() == 0 )
			throw new Exception("Could not find any cel files in the experiment directory");
		
		// add to return list, each possible cel comparison
		for( int i = 0; i < setcontrol.size(); i++)
		{
			List<File> controlFiles =  new ArrayList<File>();
			controlFiles.add( getCELFile(setcontrol.get(i),true)) ;
			
			for ( int j = 0; j < setcontrol.size(); j++)
			{
				List<File> exptFiles =  new ArrayList<File>();
				exptFiles.add(getCELFile(	setExpt.get(j),false) );
				
				returnList.add( new PincProbe(controlFiles, exptFiles) );
			}
			
		}
		
		return returnList;
	}
		
}
