package dataSetDescribers;

import java.util.ArrayList;
import java.util.List;


import parsers.GeneExperimentParser;

import utils.Median;
import utils.StatisticReturnObject;
import utils.TTest;

public class ProbeSetData implements Comparable<ProbeSetData>
{
	private final String probeSetId;
	private List<Float> baselineData;
	private List<Float> experimentData;
	private float expectedScore;
	private double calculatedScore;
	private double sortableStorage;
	private float avgDiff;
	private float foldChange;
	private float sum;
	
	public double getCalculatedScore()
	{
		return calculatedScore;
	}

	public void setCalculatedScore(double calculatedScore)
	{
		this.calculatedScore = calculatedScore;
	}

	public double getSortableStorage()
	{
		return sortableStorage;
	}
	
	public int compareTo(ProbeSetData other)
	{
		return Double.compare( this.sortableStorage, other.sortableStorage );
	}

	/*
	 *  If you set this val for every ProbeSetData in a list, these values will be used
	 *  to sort the list in ascending order. 
	 */
	public void setSortableStorage(double sortableStorage)
	{
		this.sortableStorage = sortableStorage;
	}

	public float getExpectedScore()
	{
		return expectedScore;
	}

	public void setExpectedScore(float expectedScore)
	{
		this.expectedScore = expectedScore;
	}

	public double getPooledVariance() throws Exception
	{	
		return TTest.getPooledVariance(baselineData, experimentData);	
	}
	
	/* t statistics will be reported as:
	 * 
	 *         diff 
	 *  ----------------------------------
	 *   sqrt( B*VAR + ( 1 - B) * LocalVAR ) 
	 *   
	 *   where B is the shrinkage constant
	 *   VAR is the variance across the entire array
	 *   LocalVAR is the variance of one particular gene
	 *   diff is the mean difference between conditons.
	 *   
	 *   See Box 2 in Microarray data analysis: from disarray to consolidation and consenus.  Allison et al, 
	 *   Nature Review, Genetics, Vol 7, Jan. 2006, 55-65.
	 */
	public double getShrunkenT(double overallVariance, float shrinkageConstant) throws Exception
	{
		if( shrinkageConstant < -0.01 || shrinkageConstant > 1.01)
			throw new Exception( "Shrinkage constant must be between zero and one " + shrinkageConstant );
		
			
		double top = TTest.getAverage(experimentData) - TTest.getAverage(baselineData);
		
		double bottom = shrinkageConstant * overallVariance;
		bottom += (1-shrinkageConstant) * TTest.getPooledVariance(experimentData, baselineData);
		
		return top / Math.sqrt(bottom);	
	}
	
	public StatisticReturnObject getStandardT()
	{
		return TTest.ttest(experimentData, baselineData);
	}
	
	public float calcSum() throws Exception
	{
		float sum = 0;
		
		for( Float f : baselineData )
			sum += f;
		
		for( Float f : experimentData)
			sum += f;
			
		return sum;
	}
	
	public float getBaselineAverage() throws Exception
	{
		return (float) TTest.getAverage(baselineData);
	}
	
	public float getExperimentAverage() throws Exception
	{
		return (float) TTest.getAverage(experimentData);
	}
	
	private float calcFoldChange() throws Exception
	{
		return getExperimentAverage() / getBaselineAverage();
	}
	
	public float getFoldChange() throws Exception
	{
		return foldChange;
	}
	
	public float getAverageDifference()
	{
		return avgDiff;
	}

	public float getSum()
	{
		return sum;
	}

	private float calcAverageDifference() throws Exception
	{
		return (float) (TTest.getAverage(experimentData) - TTest.getAverage(baselineData));
	}
	
	public double getFoldChangeFromMedian() throws Exception
	{
	
		float medianFoldChange = Median.getMedianVaule(experimentData) / Median.getMedianVaule(baselineData);
			
		if ( medianFoldChange < 1)
			medianFoldChange = 1 / medianFoldChange;
		
		return medianFoldChange;
	}
	
	public double getT() throws Exception
	{
		return TTest.ttestT( experimentData, baselineData);	
	}
		
	public String getProbeSetId()
	{
		return probeSetId;
	}

	private ProbeSetData( String probeSetId )
	{
		this.probeSetId = probeSetId;
	}
	
	private static boolean includeLine( String key, RatioExperiment rei ) throws Exception
	{
		AbstractSpikedControls spikeDetails = rei.getSpikeDetails();
		DataScope dataScope = rei.getDataScope();
		
		if( spikeDetails == null ) 
		{
			if( key.toUpperCase().startsWith("AFFX") )
				return false;
			
			if( dataScope.equals(DataScope.ALL) || dataScope.equals(dataScope.NULLS_ONLY) )
				return true;
			else
				return false;
		}
		
		if( spikeDetails.isExcludedProbeSet(key) )
			return false;
		
		if( dataScope.equals(DataScope.ALL) )
			return true;
		
		if( dataScope.equals(dataScope.NULLS_ONLY) )
		{
			if( spikeDetails.getAllSpikedGeneIds().contains( key ) )
				return false;
			else 
				return true;
		}
		
		// still here dataScope = SPIKES
		if( ! dataScope.equals(dataScope.SPIKES_ONLY))
			throw new Exception("Logic error");
		
		if( spikeDetails.getAllSpikedGeneIds().contains( key ) )
			return true;
		else 
			return false;
	}
	
	
	public static ExperimentData getListFromRMA( RatioExperiment rei) 
					throws Exception
	{
		ExperimentData returnVal = new ExperimentData();
		
		GeneExperimentParser gep = rei.getRMAResults();
		
		for( String key : gep.getGeneIds() )
			if( includeLine( key, rei ) )
			{
				ProbeSetData probeSetData = new ProbeSetData(key);
				probeSetData.baselineData = new ArrayList<Float>();
				probeSetData.experimentData = new ArrayList<Float>();
				
				for( String filename : rei.getBaselineFiles() )
					probeSetData.baselineData.add( gep.getDataOrThrow(key, filename) );
					
				for( String filename : rei.getExperimentFiles() )
					probeSetData.experimentData.add( gep.getDataOrThrow(key, filename) );
				
				// these are used often enough to cache
				probeSetData.sum = probeSetData.calcSum();
				probeSetData.avgDiff = probeSetData.calcAverageDifference();
				probeSetData.foldChange = probeSetData.calcFoldChange();
				
				returnVal.addProbeSetData(probeSetData);
			}
		
		return returnVal;
	}

	public List<Float> getBaselineData()
	{
		return baselineData;
	}

	public List<Float> getExperimentData()
	{
		return experimentData;
	}

	
}
