
package parsers;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Random;
import java.util.StringTokenizer;

import utils.Median;

public class CdfFileLine
{
    private final int x;
    private final int y;
    private final String qual;
    private int numInSeries;
    private final boolean isMatch;
    

    public static void checkListContains(List<CdfFileLine> list, int x, int y, boolean isMatch) throws Exception
    {
    		for ( CdfFileLine fileLine : list)
    			if ( fileLine.getX() == x && fileLine.getY() == y && fileLine.isMatch() == isMatch)
    				return;
    				
    		throw new Exception("Could not find " + x + " " + y + " " + isMatch);
    }
    
    public static HashMap<String, CdfFileLine> listToXYKeyedMap( List<CdfFileLine> list ) throws Exception
    {
    		HashMap<String, CdfFileLine> map = new HashMap<String, CdfFileLine>();
    	
    		for( CdfFileLine fileLine : list )
    		{
    			String xyKey = fileLine.getXYKey();
    			if( map.containsKey( xyKey )) 
    				throw new Exception("Duplicate key " + xyKey);
    			
    			map.put( xyKey, fileLine );
    		}
    		
    		return map;
    }

    public String getXYKey()
    {
    		return getX() + "_" + getY();
    }
    
    public CdfFileLine(String inString ) throws Exception
    {
        //System.out.println(inString);
        StringTokenizer sToken = new StringTokenizer(inString);
        
        String aToken = sToken.nextToken();
        aToken = aToken.substring(aToken.indexOf("=")+1, aToken.length());
        this.x = Integer.parseInt(aToken);
        this.y = Integer.parseInt(sToken.nextToken());
        
        sToken.nextToken();  // Probe
        sToken.nextToken();  // FEAT
        this.qual = sToken.nextToken();
        this.numInSeries = Integer.parseInt(sToken.nextToken());
        
        sToken.nextToken();  // pos
        String cbase = sToken.nextToken();  // cbase
        
        if ( cbase.length() != 1)
            throw new Exception("Parsing error.  Expecting one char for " + cbase );
        
        String pbase = sToken.nextToken();
        
        if ( pbase.length() != 1)
            throw new Exception("Parsing error.  Expecting one char for " + pbase );
        
        String tbase = sToken.nextToken();
        
        if ( tbase.length() != 1)
            throw new Exception("Parsing error.  Expecting one char for " + tbase );
        
        // as far as I can work the Affy logic, pbase == tbase means that base-pairing can't happen
        // so we have the mismatch set
        if ( pbase.equals(tbase))
            isMatch = false;
        else
            isMatch = true;
    }
    
    public boolean isMatch()
    {
        return isMatch;
    }
    public int getNumInSeries()
    {
        return numInSeries;
    }
    public String getQual()
    {
        return qual;
    }
    public int getX()
    {
        return x;
    }
    public int getY()
    {
        return y;
    }
    
    public static float getMeanOfPmAndMM(List<CdfFileLine> list, float[][] a)
    {
        float sum = 0;
        int n = 0;
        
        for ( Iterator<CdfFileLine> i = list.iterator(); i.hasNext(); )
        {
            CdfFileLine fileLine = (CdfFileLine) i.next();
            sum += a[fileLine.getX()][fileLine.getY()];
            n++;
        }
     
        return sum / n;
    }
    
    public static float getPmMedian(List<CdfFileLine> list, float[][] a) throws Exception
    {
    		
    		List<Float> vals = new ArrayList<Float>();
    		
    		for ( CdfFileLine fileLine : list )
    		{
    			if ( fileLine.isMatch())
    			{
    				vals.add(a[fileLine.getX()][fileLine.getY()]);
    			}
    		}
    		
    		Collections.sort(vals);
    		
    		if ( vals.size() == 0 )
    			throw new Exception("No PMs");
    		
    		if ( vals.size() == 1)
    			return vals.get(0);
    		
    		return Median.getMedianVaule(vals);
    		
    }
    
    public static float getPmMean(List<CdfFileLine> list, float[][] a)
    {
        float sum = 0;
        int n = 0;
        
        for ( Iterator<CdfFileLine> i = list.iterator(); i.hasNext(); )
        {
            CdfFileLine fileLine = (CdfFileLine) i.next();
            
            if ( fileLine.isMatch())
	        {
	            sum += a[fileLine.getX()][fileLine.getY()];
	            n++;
            }
        }
     
        return sum / n;
    }
    
    public static HashMap<String, List<CdfFileLine>> 
    				whittleToConstantNumberOfPms( HashMap<String, List<CdfFileLine>> map, int numberOfPms )
    		throws Exception
    {
    		HashMap<String, List<CdfFileLine>> newMap =  new HashMap<String, List<CdfFileLine>>();
    	
    		for ( String key : map.keySet())
    		{
    			List<CdfFileLine> oldList = map.get(key);
    			
    			if ( oldList.size() / 2 >= numberOfPms)
    			{
    				List<CdfFileLine> newList = new ArrayList<CdfFileLine>();
        			
        			int x=0;
        			while( newList.size() < numberOfPms * 2)
        			{
        				CdfFileLine fileLine = oldList.get(x);
        				
        				if( fileLine.isMatch )
        				{
        					newList.add( fileLine );
        					newList.add( fileLine.getCorrespondingMismatch(oldList) );
        				}
        				
        				x++;
        			}
        			
        			if ( newList.size() != numberOfPms * 2)
        				throw new Exception("Logic error " + newList.size() + " " + numberOfPms * 2 );
        			
        			newMap.put(key, newList );
    			}
    				
    		}
    		
    		return newMap;
    }
    
    /**
     *  Assign random PM-MM pairs to each key.  
     *
     */
    public static HashMap<String, List<CdfFileLine>> getScrambledMap(HashMap<String, List<CdfFileLine>> oldMap, 
    						HashSet<String> excludedGenes, Random random) throws Exception
    {
    		HashMap<String, List<CdfFileLine>> map = new HashMap<String, List<CdfFileLine>>();
    		
    		List<String> keyList = new ArrayList<String>(oldMap.keySet());
    		
    		Collections.shuffle(keyList, random);
    		
    		for ( String key : oldMap.keySet()) if ( ! excludedGenes.contains(key) )
    		{
    			List<CdfFileLine> oldCdfList = oldMap.get(key);
    			List<CdfFileLine> newCdfList = new ArrayList<CdfFileLine>();
    			for ( int y=0; y < oldCdfList.size() / 2 ; y++)
    			{
    				String randomKey = keyList.get(random.nextInt(keyList.size()));
    				List<CdfFileLine> randomCdfList = oldMap.get(randomKey);
    				CdfFileLine randomLine = randomCdfList.get( random.nextInt(randomCdfList.size() ));
    				
    				newCdfList.add(randomLine);
    				
    				if ( randomLine.isMatch)
    					newCdfList.add(randomLine.getCorrespondingMismatch(randomCdfList));
    				else
    					newCdfList.add(randomLine.getCorrespondingMatch(randomCdfList));
    			}
    			
    			map.put(key,newCdfList);
    		}
    		
    		return map;
    }
    
    private static HashMap<File, HashMap<String, List<CdfFileLine>>> cdfCacheMap = new
    			HashMap<File, HashMap<String, List<CdfFileLine>>>();
    
    /**
     *   The key is the probeSetID.  The value is a List of cdfFileLines.
     *   Each line representing a PM or MM 25 mer.
     *   
     *   Not thread safe.  Keeps an intenral cache of previously parsed cel files
     */
    public static HashMap<String, List<CdfFileLine>> parseFileForAll(File file, String chipId) throws Exception
    {
    	if(  cdfCacheMap.containsKey(file) )
    		return cdfCacheMap.get(file);
    	
        HashMap<String, List<CdfFileLine>> map = new HashMap<String, List<CdfFileLine>>();
        
        BufferedReader reader = new BufferedReader(new FileReader(file));
        
        String nextLine = reader.readLine();
        
        while (nextLine != null)
        {
            while ( nextLine != null && ! nextLine.startsWith("Name=")  )
            {
                nextLine = reader.readLine();
                //System.out.println(nextLine);
            }
                
            if ( nextLine != null)
            {
                String probeId = nextLine.substring(5);
	                
	            if ( ! probeId.equals("NONE") && ! probeId.equals(chipId))
	            {
	                //System.out.println("Got " + probeId + " from " + nextLine);
	                map.put( probeId, readListFromPositionedReader(reader, probeId));
	            }
	            
	            nextLine = reader.readLine();
            }
        }
        
        reader.close();
        
        cdfCacheMap.put(file,map);
        
        return map;
    }
    
    /**
     *  the reader should have just read the Name= line
     */
    private static List<CdfFileLine> readListFromPositionedReader(BufferedReader reader, String id) throws Exception
    {
        List<CdfFileLine> list = new ArrayList<CdfFileLine>();

        for ( int x=1; x<=6; x++)
            reader.readLine();
        
        String nextLine = reader.readLine();
        
        while ( nextLine.trim().length() > 0 )
        {
            CdfFileLine cdfl =new CdfFileLine( nextLine);
            if (! cdfl.getQual().equals(id) )
                throw new Exception("Parsing error for "+ nextLine + " expecting " + id);
            list.add( cdfl );
            
            nextLine = reader.readLine();
        }
        
        return list;
    }
    
    public static boolean hasNumProbes(List<CdfFileLine> list, int numToCheck, boolean isMatch) 
    {
        boolean[] checks = new boolean[numToCheck];
        
        for ( int x=0; x < checks.length; x++ )
            checks[x] = false;
        
        for ( int x=0; x < list.size(); x++)
        {
            CdfFileLine fileLine = (CdfFileLine) list.get(x);
            
            if ( fileLine.isMatch() == isMatch && fileLine.getNumInSeries() < numToCheck)
            {
                checks[fileLine.getNumInSeries()] = true;
            }
        }
        
        for ( int x=0; x < checks.length; x++ )
        {
            if ( ! checks[x])
                return false;
        }
        
        return true;
    }
    
    public static CdfFileLine pullOutFirst(List<CdfFileLine> list, int positionInSeries, boolean isMatch)
    {
        for ( int x=0; x < list.size(); x++)
        {
            CdfFileLine fileLine = (CdfFileLine) list.get(x);
            
            if ( fileLine.getNumInSeries() == positionInSeries && fileLine.isMatch == isMatch )
                return fileLine;
        }
        
        return null;
    }
    
    public CdfFileLine getCorrespondingMismatch(List<CdfFileLine> cdfList) throws Exception
    {
    		if( ! isMatch )
    			throw new Exception("Logic error");
    		
    		for( CdfFileLine fileLine : cdfList)
    		{
    			
    			if( fileLine.getX() == getX() && fileLine.getY() == getY() +1 )
    			{
    				if ( fileLine.isMatch() )
    					throw new Exception("Parsing error. Expecting mismatch");
    				
    				return fileLine;
    			}
    		}
    		
    		throw new Exception("Could not find mismatch");
    }
    
    public CdfFileLine getCorrespondingMatch(List<CdfFileLine> cdfList) throws Exception
    {
    		if( isMatch )
    			throw new Exception("Logic error");
    		
    		for( CdfFileLine fileLine : cdfList)
    			if( fileLine.getX() == getX() && fileLine.getY()  == getY() -1 )
    			{
    				if (! fileLine.isMatch() )
    					throw new Exception("Parsing error. Expecting match");
    				
    				return fileLine;
    			}
    		
    		throw new Exception("Could not find match");
    }
    
    public static List<CdfFileLine> getForId(File file, String id) throws Exception
    {
        BufferedReader reader = new BufferedReader(new FileReader(file));
        
        String nextLine = reader.readLine();
        
        String matchString = "Name=" + id;
        System.out.println("Searching for " + matchString);
        while ( ! nextLine.startsWith(matchString))
            nextLine = reader.readLine();
        
        List<CdfFileLine> list = readListFromPositionedReader(reader, id);
        
        reader.close();
        return list;
    }
    
    public static List<CdfFileLine> getPMLines (List<CdfFileLine> list)
    {
    	List<CdfFileLine> returnList = new ArrayList<CdfFileLine>();
    	
    	for( CdfFileLine line : list )
    		if( line.isMatch())
    			returnList.add(line);
    	
    	return returnList;
    }
    
    
}
