package parsers;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;

import dataSetDescribers.FilePathGenerator;
import dataSetDescribers.RatioExperiment;

import statistics.StatisticsProvider;
import utils.KS_Test;
import utils.Pearson;

public class FdrPermutationFileLine
{
	private String affyId;
	private boolean isTruePositive;
	private float actualFdr;
	private float predictedFdr;
	
	private FdrPermutationFileLine(String fileLine) throws Exception
	{
		StringTokenizer sToken = new StringTokenizer(fileLine);
		
		this.affyId = sToken.nextToken();
		this.isTruePositive = Boolean.parseBoolean(sToken.nextToken());
		this.actualFdr = Float.parseFloat(sToken.nextToken());
		this.predictedFdr = Float.parseFloat(sToken.nextToken());
		
		if( sToken.hasMoreTokens() ) 
			throw new Exception("Unexpected token " + sToken.nextToken());
	}
	
	public static List<FdrPermutationFileLine> parseFile(File file) throws Exception
	{
		List<FdrPermutationFileLine> list = new ArrayList<FdrPermutationFileLine>();
		
		BufferedReader reader = new BufferedReader(new FileReader(file));
		
		reader.readLine();
		
		String nextLine = reader.readLine();
		
		while( nextLine != null)
		{
			list.add(new FdrPermutationFileLine(nextLine));
			nextLine = reader.readLine();
		}
			
		reader.close();
		return list;
	}
	
	public static int getNumPredictedBelow( StatisticsProvider sp, RatioExperiment re, float cutoff ) throws Exception
	{
		List<FdrPermutationFileLine> list = parseFile(FilePathGenerator.getFDRFile(re, sp, true));
		
		int returnNum=0;
		
		for( FdrPermutationFileLine fdr : list )
		{	
			if( fdr.getPredictedFdr() <= cutoff )
				returnNum++;
			else return returnNum;
		}
			
		return returnNum;
	}
	
	public float getActualFdr()
	{
		return actualFdr;
	}

	public String getAffyId()
	{
		return affyId;
	}

	public boolean isTruePositive()
	{
		return isTruePositive;
	}

	public float getPredictedFdr()
	{
		return predictedFdr;
	}
	
	public static double getKsVal(StatisticsProvider sp, RatioExperiment re) throws Exception
	{
		File fdrFile = FilePathGenerator.getFDRFile(re, sp, true);
		
		List<FdrPermutationFileLine> list = parseFile(fdrFile);
		List<Double> l1 = new ArrayList<Double>();
		List<Double> l2 = new ArrayList<Double>();
		
		for( FdrPermutationFileLine line : list )
		{
			l1.add( new Double( line.getPredictedFdr()));
			l2.add( new Double( line.getActualFdr()));
		}
		
		return KS_Test.getKS_P_Value(l1, l2);
	}
	
	public static double getPearsonRSquared(StatisticsProvider sp, RatioExperiment re) throws Exception
	{
		File fdrFile = FilePathGenerator.getFDRFile(re, sp, true);
		
		List<FdrPermutationFileLine> list = parseFile(fdrFile);
		List<Number> l1 = new ArrayList<Number>();
		List<Number> l2 = new ArrayList<Number>();
		
		for( FdrPermutationFileLine line : list )
		{
			l1.add( new Double( line.getPredictedFdr()));
			l2.add( new Double( line.getActualFdr()));
		}
		
		double rVal = Pearson.getPearsonR(l1, l2);
		return rVal * rVal;
	}
	
	public static SensSpecReturnObject getSensSpecAtFdr( StatisticsProvider sp, RatioExperiment re, float cutoff ) 
		throws Exception
	{
		
		float maxSoFar= Float.MIN_VALUE;
		File fdrFile = FilePathGenerator.getFDRFile(re, sp, true);
		List<FdrPermutationFileLine> list = parseFile(fdrFile);
		
		float numCalls = 0;
		float numCorrect = 0;
		float totalNumTruePositives = 0;
		
		for( FdrPermutationFileLine line : list )
		{
			if( line.isTruePositive() )
				totalNumTruePositives++;
			
			maxSoFar = Math.max( maxSoFar, line.getPredictedFdr() );
			
			if( maxSoFar <= cutoff  )
			{
				numCalls++;
				
				if( line.isTruePositive() )
					numCorrect++;
			}
		}
		
		if( numCalls == 0 )
			return new SensSpecReturnObject(0,0);
		
		float sensitivity = numCorrect / totalNumTruePositives;
		float spec = numCorrect / numCalls;
		
		return new SensSpecReturnObject(sensitivity,spec);
	}
}
