package parsers;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.StringTokenizer;


import scripts.ProbeMeanVsSD;
import utils.RandomNumberGenerator;

/*
 *  Parses a file of the format
 *  
 *  /tExperiment1/tExperiment2/t....
 *  geneName\tdata\tdata\t....
 *  geneName\tdata\tdata\t....
 *  
 *  This implementation is a bit of a hack and is very RAM inefficient.
 *  In addition, it will fail if the string "@!@" appears anywhere in the input file
 *  
 */
public class GeneExperimentParser
{
	private static final String SEPARATION_STRING = "@!@";
	
	private HashMap<String,Float> data;
	
	public float getDataOrThrow(String geneId, String experimentName) throws Exception
	{
		String key = geneId + SEPARATION_STRING + experimentName;
		
		Float returnVal = data.get(key);
		
		if( returnVal == null )
			throw new Exception("Could not find " +key);
		
		return returnVal;
	}
	
	public void scrambleData()
	{
		List<String> oldKeys = new ArrayList<String>( data.keySet());
		List<Float> oldValues = new ArrayList<Float>( data.values());
		
		data = new HashMap<String,Float>();
		
		Collections.shuffle(oldKeys);
		
		for( int x=0; x < oldKeys.size(); x++ )
			data.put( oldKeys.get(x), oldValues.get(x) );
	}
	
	public HashSet<String> getGeneIds()
	{
		HashSet<String> returnSet = new HashSet<String>();
		
		for( String key : data.keySet() )
		{
			returnSet.add( key.substring( 0, key.indexOf(SEPARATION_STRING)));
		}
			
		return returnSet;
	}
	
	public HashSet<String> getExperimentNames()
	{
		HashSet<String> returnSet = new HashSet<String>();
		
		for( String key : data.keySet() )
		{
			returnSet.add( key.substring( key.indexOf(SEPARATION_STRING) 
					+ SEPARATION_STRING.length() ));
		}
			
		return returnSet;
	}
	
	/*
	 *  Fills with random data from the standard normal distribution.
	 */
	public GeneExperimentParser(List<String> baselineFiles, List<String> experimentFiles, int sampleSize, int numGenes)
		throws Exception
	{
		RandomNumberGenerator rng = new RandomNumberGenerator();
		
		this.data = new HashMap<String,Float>();
		
		for( int x=0; x < numGenes; x++ )
		{
			String geneId = "gene_" + x;
			
			for( String s : baselineFiles )
			{
				addData(geneId, s, (float)rng.nextGaussian());
			}
			
			for( String s :experimentFiles )
			{
				addData(geneId, s, (float)rng.nextGaussian());
			}
		}
	}
	
	public void addData( String geneId, String fileName, float dataVal ) throws Exception
	{
		String key = geneId + SEPARATION_STRING + fileName;
		
		if(  data.get(key) != null )
			throw new Exception("Duplicate key " + key);
		
		data.put(key, dataVal);
	}
	
	/*
	 * Generates data comparing two files at the PROBE level
	 */
	public GeneExperimentParser( File cdfFile,
								 String chipId,
								File baselineFile, 
								File experimentFile,
								int expectedNumPMProbes,
								String baselinePrefix,
								String experimentPrefix) throws Exception
	{
		this.data = new HashMap<String,Float>();
		HashMap<String, List<CdfFileLine>> map = CdfFileLine.parseFileForAll(
				cdfFile, chipId);
		float[][] baselineVals = CelParserFileLine.parseFileForArray(baselineFile);
		float[][] expVals = CelParserFileLine.parseFileForArray(experimentFile);
		
		populateMap(map, expectedNumPMProbes, baselinePrefix, experimentPrefix, baselineVals, expVals);
	}
	
	/*
	 * Generates data comparing two files at the PROBE level.
	 * Returns log(2) data
	 */
	public GeneExperimentParser( File cdfFile,
								 String chipId,
								List<File> baselineFiles, 
								List<File> experimentFiles,
								int expectedNumPMProbes,
								String baselinePrefix,
								String experimentPrefix) throws Exception
	{
		this.data = new HashMap<String,Float>();
		HashMap<String, List<CdfFileLine>> map = CdfFileLine.parseFileForAll(
				cdfFile, chipId);
		float[][] baselineVals = CelParserFileLine.parseFilesForAverage(baselineFiles);
		float[][] expVals = CelParserFileLine.parseFilesForAverage(experimentFiles);
		populateMap(map, expectedNumPMProbes, baselinePrefix, experimentPrefix, baselineVals, expVals);
	}
	
	private void populateMap(HashMap<String, List<CdfFileLine>> map, int expectedNumPMProbes, 
			String baselinePrefix, String experimentPrefix, float[][] baselineVals, float[][] expVals)
		throws Exception
	{

		for( String key : map.keySet()  )
		{
			List<CdfFileLine> pmList = CdfFileLine.getPMLines( map.get(key));
			
			if( pmList.size() >= expectedNumPMProbes )
			{
				int index = 0;
    			
    			
    			while( index < expectedNumPMProbes)
    			{
    				CdfFileLine pmLine = pmList.get(index);
    				
    				addData(key, baselinePrefix + index,
    						(float) ProbeMeanVsSD.log2(baselineVals[pmLine.getX()][pmLine.getY()]));
    				
    				addData(key, experimentPrefix + index, 
    						(float) ProbeMeanVsSD.log2(expVals[pmLine.getX()][pmLine.getY()]));
    				index++;
    			}
			}
			else
			{
				System.out.println("Skipping " + key + " only " + pmList.size() + " PM probes");
			}
		}
	}
	
	public GeneExperimentParser(File rmaFile) throws Exception
	{
		this.data = new HashMap<String,Float>();
		List<String> topLine = new ArrayList<String>();
		BufferedReader reader = new BufferedReader(new FileReader(rmaFile));
		
		StringTokenizer sToken = new StringTokenizer(reader.readLine(),"\t");
	
		// ignore the very first token
		sToken.nextToken();
		
		while ( sToken.hasMoreTokens())
		{
			String fileName = sToken.nextToken();
			topLine.add( fileName );
		}
		
		String nextLine = reader.readLine();
		
		while(nextLine != null)
		{
			sToken = new StringTokenizer(nextLine);
			
			int index =0;
			String geneId = sToken.nextToken();
			
			while( sToken.hasMoreTokens())
			{
				String key = geneId + SEPARATION_STRING + topLine.get(index);
				index++;
				
				if(  data.get(key) != null )
					throw new Exception("Duplicate key " + key);
				
				data.put( key, java.lang.Float.parseFloat(sToken.nextToken()) );
			}
			
			nextLine = reader.readLine();
		}
		
		reader.close();
	}
	
	public GeneExperimentParser()
	{
		this.data = new HashMap<String, Float>();
	}
}
