package parsers;

import java.io.BufferedInputStream;
import java.io.BufferedWriter;
import java.io.DataInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;



import dataSetDescribers.FilePathGenerator;
import dataSetDescribers.LatinSquareGeneric;
import dataSetDescribers.RatioExperiment;

import statistics.ShrunkenVariance;
import statistics.StatisticsLevelNormalization;
import statistics.StatisticsProvider;
import utils.TTest;

public class ScoreFileLineParser implements Comparable<ScoreFileLineParser>
{
	private String probeSetId;
	private boolean isTruePositive;
	private float average;
	private float score;
	private float unnormalizedFoldChange;
	private double pValue;
	private double expectedPValue;
	private float expectedScore;
	private float unnormalizedScore;
	private double bhFDR;
	private double byFDR;
	
	
	
	@Override
	public boolean equals(Object obj)
	{
		ScoreFileLineParser other= (ScoreFileLineParser) obj;
		return this.probeSetId.equals(other.getProbeSetId());
	}

	@Override
	public int hashCode()
	{
		return this.getProbeSetId().hashCode();
	}

	public ScoreFileLineParser(float score, float average)
	{
		this.score = score;
		this.average = average;
	}
	
	public ScoreFileLineParser(float score, String probeSetID)
	{
		this.score = score;
		this.probeSetId = probeSetID;
		this.isTruePositive = false;
	}
	
	public ScoreFileLineParser(Double score, String probeSetID)
	{
		this.score = score.floatValue();
		this.probeSetId = probeSetID;
		this.isTruePositive = false;
	}
	
	private ScoreFileLineParser(DataInputStream in) throws Exception
	{
		this.probeSetId = in.readUTF();
		this.isTruePositive = in.readBoolean();
		this.average = in.readFloat();
		this.score = (float) in.readDouble();
		this.unnormalizedFoldChange =(float) in.readDouble();
		this.pValue = in.readDouble();
		this.expectedPValue = in.readDouble();
		this.expectedScore = in.readFloat();
		this.unnormalizedScore = in.readFloat();
		this.bhFDR = in.readDouble();
		this.byFDR = in.readDouble();
	}
	
	
	public static void replaceScoresWithNormalDistribution( List<ScoreFileLineParser> list ) 
		throws Exception
		{

			List<Float> scores = new ArrayList<Float>();
			
			for( int x=0; x < list.size(); x++)
			{
				ScoreFileLineParser aLine = list.get(x);
				scores.add( aLine.getScore() );
			}
			
			double mu = TTest.getAverage(scores);
			double sd = TTest.getStDev(scores);
		
			for( ScoreFileLineParser sflp : list )
			{
				sflp.score = (float) (Math.abs(  sflp.getScore()  - mu) / sd);
			}
		}
	
	public static List<Float> getScores(RatioExperiment re, StatisticsProvider sp) throws Exception
	{
		List<ScoreFileLineParser> list = parseFile(re,sp);
		
		List<Float> returnList = new ArrayList<Float>();
		
		for( ScoreFileLineParser sflp : list) 
			returnList.add(sflp.getScore());
		
		return returnList;
	}
	
	private void writeThisLine(BufferedWriter writer) throws Exception
	{
		writer.write( this.probeSetId + "\t" );
		writer.write( this.isTruePositive + "\t");
		writer.write( this.average+ "\t");
		writer.write(this.score+ "\t");
		writer.write(this.unnormalizedFoldChange + "\t");
		writer.write(this.pValue + "\t");
		writer.write(this.expectedPValue + "\t");
		writer.write(this.expectedScore+ "\t");
		writer.write(this.unnormalizedScore + "\t");
		writer.write(this.bhFDR + "\t");
		writer.write(this.byFDR + "\n");
	}
	

	public static void convertFromBinaryFile(RatioExperiment re, 
			StatisticsProvider sp, boolean addLowessLine, boolean statNormalizeList) throws Exception
	{
		List<ScoreFileLineParser> list = ScoreFileLineParser.parseFile(re,sp);
		
		if( addLowessLine )
			addLoessLineToExpectedScore(list);
		
		if( statNormalizeList)
		{
			Collections.sort( list, new SortByAverage() );
			StatisticsLevelNormalization.normalizeList(list, true);
			
			for ( int x=0; x < list.size(); x++)
			{
				ScoreFileLineParser sflp = list.get(x);
				sflp.expectedPValue = (x+1.0) / list.size();
			}
		}
			
		
		String oldFilePath = FilePathGenerator.getScoreFile(re,sp).getAbsolutePath();
		String newFilePath= oldFilePath.substring(0, oldFilePath.lastIndexOf(".")) + ".txt";
		
		BufferedWriter writer = new BufferedWriter(new FileWriter(new File( newFilePath )));
		writeScoreHeader(writer,re,sp);
		
		for( ScoreFileLineParser sflp : list )
			sflp.writeThisLine(writer);
		
		writer.flush();  writer.close();
		
	}
	
	public static void main(String[] args) throws Exception
	{
		RatioExperiment re = new LatinSquareGeneric(8,9);
		re.setUseQuantileQuantileNormalization(true);
		re.setUseStatisticLevelNormalization(false);
		StatisticsProvider sp = new ShrunkenVariance(1.0f);
		convertFromBinaryFile(re, sp, false, false);
		
		/*
		RatioExperiment re = new Eb24TreatmentVsControl();
		re.setUseQuantileQuantileNormalization(true);
		re.setUseStatisticLevelNormalization(false);
		re.setDataScope(DataScope.ALL);
		StatisticsProvider sp = new CyberT();
		
		convertFromBinaryFile(re,sp,false,false);
		System.out.println("Converted " + FilePathGenerator.getScoreFile(re,sp));
		
		/*
		List<Integer> baselineInts = new ArrayList<Integer>();
		baselineInts.add(13); baselineInts.add(14); baselineInts.add(16);
		
		List<Integer> experimentInts = new ArrayList<Integer>();
		experimentInts.add(15); experimentInts.add(17); experimentInts.add(18);
		
		RatioExperiment re = new AAmprnaGeneric("Amp_TB",baselineInts, experimentInts);
		re.setUseQuantileQuantileNormalization(true);
		re.setUseStatisticLevelNormalization(true);
		re.setDataScope(DataScope.ALL);
		StatisticsProvider sp = new CyberT();
		*/
		
		/*
		RatioExperiment re = new Aldrich_FourScns();
		StatisticsProvider sp = new CyberT();
		re.setUseQuantileQuantileNormalization(true);
		re.setUseStatisticLevelNormalization(false);
		re.setDataScope(DataScope.ALL);
		*/
	}
	
	public static String getColumnHeader( RatioExperiment rei, StatisticsProvider sp )
		throws Exception
	{
		return rei.getShortExperimentName() + "_" + 
		sp.getDescription() + "_" + rei.getDataScope().getDescription(); 
		
	}
	
	public static void writeScoreHeader(BufferedWriter writer, RatioExperiment rei, StatisticsProvider sp) 
				throws Exception
	{
		String columnHeader = getColumnHeader(rei,sp);
		
		
//		 pValue, expectedDiff and unNormalizedScore will not be defined for all statistics
		writer.write("ProbeSetID_" + columnHeader +  "\t" +  
			  	"isTruePositive_" + columnHeader + "\t" + 
			  	 "average_" + columnHeader + "\t" +  
			  	  "Score_" + columnHeader +  "\t" +  
			  	  "ApproximateFoldChange_" + columnHeader + "\t" + 
			  	  "pValue_" + columnHeader +  "\t" + 
			  	  "expectedPValue_" + columnHeader + "\t" + 
			  	  "expectedScore_" +  columnHeader + "\t" +  
			  	  "unNormalizedScore_" + columnHeader  +"\t" +
			  	  "BH_FDR_" + columnHeader + "\t" + 
			  	  "BY_FDR_" + columnHeader + "\n"  );
	}
	
	public void setUnnormalizedScore(float unnormalizedScore)
	{
		this.unnormalizedScore = unnormalizedScore;
	}

	public static void addLoessLineToExpectedScore( List<ScoreFileLineParser> list) throws Exception
	{

			// Code no longer available for this project
		
	}


	public static NumTrueAndFalsePositives getPositiveCount( RatioExperiment rei, 
			StatisticsProvider sp, 
			float cutoff, 
			boolean useBY) throws Exception
	{
		File scoreFile = FilePathGenerator.getScoreFile(rei,sp);
		int numTruePositive =0;
		int numFalsePositive =0 ;
		
		List<ScoreFileLineParser> list = parseFile(scoreFile);
		
		for( ScoreFileLineParser line : list )
		{
			boolean add = false;
			
			if( useBY  )
			{
				if( line.getByFDR() <= cutoff )
					add=true;
			} 
			else
			{
				if( line.getBhFDR() <= cutoff)
					add = true;
			}
			
			if( add )
			{
				if( rei.getSpikeDetails() == null  )
					numFalsePositive++;
				else if( rei.getSpikeDetails().getAllSpikedGeneIds().contains( line.getProbeSetId() ) )
					numTruePositive++;
				else
					numFalsePositive++;
			}
		}
		
		return new NumTrueAndFalsePositives(numTruePositive, numFalsePositive);
	}
	
	public static double getPValueRMSD( RatioExperiment rei, StatisticsProvider sp ) 
		throws Exception
		{
			double sum = 0;
			int n=0;
			System.out.println( FilePathGenerator.getScoreFile(rei,sp).getAbsolutePath()  );
			List<ScoreFileLineParser> list = parseFile( FilePathGenerator.getScoreFile(rei,sp) );
			
			for ( ScoreFileLineParser fileLine : list)
			{
				double val = fileLine.pValue - fileLine.expectedPValue;
				sum += val * val;
				n++;
			}
			
			return 1000 * Math.sqrt(sum / n);
		}
	
	public static HashSet<ScoreFileLineParser> getAsSet( RatioExperiment rei, StatisticsProvider sp)
		throws Exception
	{
		HashSet<ScoreFileLineParser> set = new HashSet<ScoreFileLineParser>();
		
		List<ScoreFileLineParser> list = parseFile(rei, sp);
		
		for( ScoreFileLineParser sflp : list )
		{
			if( set.contains(sflp))
				throw new Exception("Found duplicate ");
			
			set.add(sflp);
		}
		
		return set;
	}
	
	public static HashSet<String> getKeysForTopScores( File fileToParse, int numToGet ) throws Exception
	{
		List<ScoreFileLineParser> list = parseFile(fileToParse);
		
		HashSet<String> set=  new HashSet<String>();
		
		for( int x=0; x < numToGet; x++ )
			set.add( list.get(x).getProbeSetId() );
		
		return set;
		
	}
	
	
	public static List<String> getKeys( List<ScoreFileLineParser> list ) throws Exception
	{
		
		List<String> returnList=  new ArrayList<String>();
		
		for( int x=0; x < list.size(); x++ )
			returnList.add( list.get(x).getProbeSetId() );
		
		return returnList;
		
	}
	
	public static List<String> getKeysForCutoff(RatioExperiment rei, 
			StatisticsProvider sp , float cutoff, boolean useBh ) throws Exception
	{
		File file = FilePathGenerator.getScoreFile(rei,sp);
		return getKeysForCutoff(file, cutoff, useBh);
	}
	
	public static List<String> getKeysForCutoff( File fileToParse, 
			float cutoff, boolean useBh  ) throws Exception
	{
		List<ScoreFileLineParser> list = parseFile(fileToParse);
		
		List<String> returnList=  new ArrayList<String>();
		
		for( ScoreFileLineParser sflp : list)
		{
			if( useBh && sflp.getBhFDR() < cutoff )
				returnList.add( sflp.getProbeSetId() );
			
			if( !useBh && sflp.getByFDR() < cutoff)
				returnList.add( sflp.getProbeSetId() );
		}
			
		return returnList;
		
	}
	
	/*
	 *   The key is the probeSetId
	 */
	public static HashMap<String, ScoreFileLineParser > getAsMap(File fileToParse) throws Exception
	{
		HashMap<String, ScoreFileLineParser> map = new HashMap<String, ScoreFileLineParser>();
		
		for( ScoreFileLineParser sfp : parseFile(fileToParse) )
		{
			if(  map.get(sfp.getProbeSetId()) != null) 
				throw new Exception("Two for " + sfp.getProbeSetId());
			
			map.put( sfp.getProbeSetId(), sfp );
		}
		
		return map;
	}
	
	public static List<ScoreFileLineParser> parseFile(RatioExperiment rei,
			StatisticsProvider sp) throws Exception
	{
		File scoreFile = FilePathGenerator.getScoreFile(rei,sp);
		return parseFile(scoreFile);
	}
	
	public static List<ScoreFileLineParser> parseFile(File fileToParse) throws Exception
	{
		List<ScoreFileLineParser> list = new ArrayList<ScoreFileLineParser>();
		
		 DataInputStream in = new DataInputStream( new BufferedInputStream(
	                		new FileInputStream(fileToParse) ));
		
		 int numLines = in.readInt();
		 
		 for( int x=0; x < numLines; x++)
		 {
			 list.add( new ScoreFileLineParser(in));
		 }
		
		in.close();
		return list;
	}
	
	
	public double getBhFDR()
	{
		return bhFDR;
	}

	public double getByFDR()
	{
		return byFDR;
	}

	public boolean isATruePositive()
	{
		return isTruePositive;
	}

	public boolean setIsATruePositive(boolean isTrue)
	{
		this.isTruePositive = isTrue;
		return isTruePositive;
	}
	public double getRocScore()
	{
		return Math.abs(score);
	}

	public boolean ignoreThisLine()
	{
		return false;
	}

	public float getExpectedScore()
	{
		return expectedScore;
	}
	
	public void setExpectedScore(float newExpectedScore)
	{
		this.expectedScore = newExpectedScore;
	}

	public boolean isTruePositive()
	{
		return isTruePositive;
	}

	public String getProbeSetId()
	{
		return probeSetId;
	}

	public double getPValue()
	{
		return pValue;
	}

	public float getScore()
	{
		return score;
	}

	public float getAverage()
	{
		return average;
	}

	public float getUnnormalizedFoldChange()
	{
		return unnormalizedFoldChange;
	}

	public float getUnnormalizedScore()
	{
		return unnormalizedScore;
	}

	public double getExpectedPValue()
	{
		return expectedPValue;
	}

	
	
	public void setBhFDR(double bhFDR)
	{
		this.bhFDR = bhFDR;
	}

	public void setByFDR(double byFDR)
	{
		this.byFDR = byFDR;
	}

	public void setExpectedPValue(double expectedPValue)
	{
		this.expectedPValue = expectedPValue;
	}

	public void setPValue(double value)
	{
		this.pValue= value;
	}

	public void setScore(float score)
	{
		this.score = score;
	}

	public int compareTo(ScoreFileLineParser other)
	{
		return Double.compare( Math.abs( other.getScore()), Math.abs( this.getScore()) );
	}
	
	public static final class SortByAverage implements Comparator<ScoreFileLineParser>
	{
		public int compare(ScoreFileLineParser sflp1, ScoreFileLineParser sflp2)
		{
			return Float.compare( sflp1.getAverage(), sflp2.getAverage()  );
		}
	}
	
	public static final class SortByPValue implements Comparator<ScoreFileLineParser>
	{
		public int compare(ScoreFileLineParser sflp1, ScoreFileLineParser sflp2)
		{
			return Double.compare( sflp1.getPValue(), sflp2.getPValue()  );
		}
	}
	
	public static final class SortByProbeId implements Comparator<ScoreFileLineParser>
	{
		public int compare(ScoreFileLineParser sflp1, ScoreFileLineParser sflp2)
		{
			return sflp1.getProbeSetId().compareTo(sflp2.getProbeSetId());
		}
	}
}
