package statistics;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

import rScripts.ShellToR;

import utils.ConfigReader;
import utils.StatisticReturnObject;
import dataSetDescribers.ExperimentData;
import dataSetDescribers.FilePathGenerator;
import dataSetDescribers.ProbeSetData;

/*
 *   This works by shelling out to R
 *   
 *   THIS ASSUMES THAT THERE ARE 11 probes.  Change probe level size if this is not the case.
 *   
 */
public class AnovaViaR implements StatisticsProvider
{
	private List<Double> fValues =new ArrayList<Double>();
	private List<Double> pValues = new ArrayList<Double>();
	private HashMap<String, Integer> positions = new HashMap<String, Integer>();
	private boolean prelimCalculationsDone = false;
	private final static int probelevelSize = 11;
	
	// for testing
	public List<Double> getPValues()
	{
		return Collections.unmodifiableList( pValues );
	}
	
	// for testing
	public List<Double> getFValues()
	{
		return Collections.unmodifiableList(fValues);
	}
	
	// for testing
	public Map<String,Integer> getPositions()
	{
		return Collections.unmodifiableMap(positions);
	}
	
	// for testing
	public HashMap<Integer, String> reversePositions() throws Exception
	{
		HashMap<Integer, String> map = new HashMap<Integer, String>();
		
		for( String key : positions.keySet() )
		{
			Integer anInt = positions.get(key);
			
			if( map.get(anInt) != null)
				throw new Exception("Dual keys " + key);
			
			map.put( anInt, key );
		}
		
		return map;
	}
	
	
	protected boolean doLowessNormalization()
	{
		return false;
	}
	
	
	private void writeCommandFile( ExperimentData experiment  ) throws Exception
	//	Parse the expt and controls and send each list of probes to R for ANONA analysis
	{
		List<ProbeSetData> listOFProbeSetData = experiment.getProbeSetData();
		
		for( int x = 0; x < listOFProbeSetData.size(); x++ )
		{
			ProbeSetData data = listOFProbeSetData.get(x);
						
			List<Float> baselineFloats = data.getBaselineData();
			List<Float> exptFloats = data.getExperimentData();
			
			//calculate Number of repeats, assuming # of probes = probelevelSize.
			if (baselineFloats.size() != exptFloats.size()) {throw new Exception("Size of list do not match");}
			
			
			int totalbits = baselineFloats.size();
			if ( (totalbits % probelevelSize)   != 0) {throw new Exception("Size of list does not match expected number of probes");}
			int numofRepeats = 0;
			while (totalbits > 0)
			{
				totalbits = totalbits - probelevelSize;
				numofRepeats++;
			}
			if  (numofRepeats < 1) {throw new Exception("Num of Repeats is Wrong, didn't get a single repeat");}
			
			makecommandFile2(baselineFloats, exptFloats, numofRepeats, data);

		}			
	}
	
	private void makecommandFile2(List<Float> baselineFloats, List<Float> exptFloats, int numofRepeats, ProbeSetData data) throws Exception
	{
		// Send this stuff to R for Anovamagic
		File commandFile = FilePathGenerator.deleteAndEnsureDeleted(
				ConfigReader.getAffyAnalysisOutputDir() + File.separator + "rAnovaCommands.R");
		
		BufferedWriter writer = new BufferedWriter(new FileWriter(commandFile));
		
		if (numofRepeats != 1) throw new Exception(" Ooops, I didn't code for repeats yet");
		if (numofRepeats == 1)
		{
			
		
			StringBuffer probeValuesString = new StringBuffer("data <- c(");
			
			
			//make probevalue
			for (Float f: baselineFloats)
			{
				probeValuesString.append(f.toString() + ",");
			}
			
			for (Float f: exptFloats)
			{
				probeValuesString.append(f.toString() + ",");
			}
			
			//remove that last comma
			probeValuesString.deleteCharAt((probeValuesString.length()-1));
			probeValuesString.append(")");  //tested, first line good to go
			
			
			
			//make experiment factor line
			StringBuffer exptString = new StringBuffer();
			exptString.append("experiment <- c(");
			for (int x = 0; x < probelevelSize; x++)
			{
				exptString.append("\"E1\",");
			}
			for (int x = 0; x < probelevelSize; x++)
			{
				exptString.append("\"E2\",");
			}
			exptString.deleteCharAt((exptString.length()-1));
			exptString.append(")");
			
			//make probe factor line
			StringBuffer probefactorString = new StringBuffer();
			probefactorString.append("probes <- c(");
			for (int x = 1; x <= probelevelSize; x++)
			{
				probefactorString.append("\"P" + x + "\",");
			}
			for (int x = 1; x <= probelevelSize; x++)
			{
				probefactorString.append("\"P" + x + "\",");
			}
			probefactorString.deleteCharAt((probefactorString.length()-1));
			probefactorString.append(")");
			
			//write the command line
			writer.append(probeValuesString + "\n");
			writer.append(exptString + "\n");
			writer.append(probefactorString + "\n");
			writer.append("mylm <- lm(data ~ experiment + probes)" + "\n");
			writer.append("x <- anova(mylm)" + "\n");
			
			File rFResults = FilePathGenerator.deleteAndEnsureDeleted( ConfigReader.getAffyAnalysisOutputDir()
					+ File.separator + "rFvalueResultOkToDelete.txt");
			File rPResults = FilePathGenerator.deleteAndEnsureDeleted( ConfigReader.getAffyAnalysisOutputDir()
					+ File.separator + "rPvalueResultOkToDelete.txt");
			
			writer.write("write.csv(x$F, file=\"" + FilePathGenerator.getRPath(rFResults)
					+ "\");\n" );
			
			writer.write("write.csv(x$Pr, file=\"" + FilePathGenerator.getRPath(rPResults)
					+ "\");\n" );
			
			writer.flush();  writer.close();
			ShellToR.shellToR(commandFile);
			parseResultsFile(rFResults, rPResults,  data);
			
		}		
	}

	private void parseResultsFile(File resultsF, File resultsP, ProbeSetData data) throws Exception
	{
		// TODO Auto-generated method stub
		System.out.println( "Parsing results file from Anova R " + resultsF.getAbsolutePath() );
		
		BufferedReader readerF = new BufferedReader(new FileReader(resultsF));
		BufferedReader readerP = new BufferedReader(new FileReader(resultsP));
		
		readerF.readLine(); //throw away first line
		readerP.readLine(); 
		
		String nextLine = readerF.readLine(); //our value of interest
		
		
			StringTokenizer sToken = new StringTokenizer(nextLine, ",");
			sToken.nextToken(); //not needed
			Double fValue = Double.valueOf(sToken.nextToken());
			
		nextLine = readerP.readLine(); //our value of interest
			
			
			sToken = new StringTokenizer(nextLine, ",");
			sToken.nextToken(); //not needed
			Double pValue = Double.valueOf(sToken.nextToken());
			
			
			
		int index = fValues.size();
			
		if (fValues.size() != index )
			throw new Exception("Parsing error");
		
		if ( pValues.size() != index)
			throw new Exception("Parsing error");
			
			
		positions.put(data.getProbeSetId(),index);
			
		fValues.add( fValue);
		
		pValues.add( pValue);

		
		readerF.close();
		readerP.close();
	}
	

	public void preliminaryCalculations(ExperimentData experiment) throws Exception
	{
		if( prelimCalculationsDone )
			throw new Exception("Preliminary Calculations should only be performed once");
		
		writeCommandFile(experiment);
		
		prelimCalculationsDone = true;
	}

	public StatisticReturnObject getStats(ProbeSetData probeSet) throws Exception
	{
		int index = positions.get(probeSet.getProbeSetId());
		
		return new StatisticReturnObject( fValues.get(index), pValues.get(index));
	}

	public boolean hasPValue()
	{
		return true;
	}

	public String getDescription()
	{
		return "ANOVA";
	}
	
	public static void main(String[] args) throws Exception
	{
		//testing a loop
		int totalbits = 121;
		if ( (totalbits % probelevelSize)   != 0) {throw new Exception("Size of list does not match expected number of probes");}
		int numofRepeats = 0;
		while (totalbits > 0)
		{
			totalbits = totalbits - probelevelSize;
			numofRepeats++;
			
		}
		if  (numofRepeats < 1) {throw new Exception("Num of Repeats is Wrong, didn't get a single repeat");}
		
		List<Float> floats = new ArrayList<Float>();
		StringBuffer probeValuesString = new StringBuffer("data <- c(");
		
		
		floats.add(3.2f);
		floats.add(45.9f);
		floats.add(9.0f);
		floats.add(1.0006f);
		
		
		for (Float f: floats)
		{
			probeValuesString.append(f.toString() + ",");
		}
		
		//remove that last comma
		probeValuesString.deleteCharAt((probeValuesString.length()-1));
		probeValuesString.append(")");
		
		System.out.println(probeValuesString);
		
		
		//test
		StringBuffer exptString = new StringBuffer();
		exptString.append("experiment <- c(");
		for (int x = 0; x < probelevelSize; x++)
		{
			exptString.append("\"E1\",");
		}
		for (int x = 0; x < probelevelSize; x++)
		{
			exptString.append("\"E2\",");
		}
		exptString.deleteCharAt((exptString.length()-1));
		exptString.append(")");
		System.out.println(exptString);
		
		StringBuffer probefactorString = new StringBuffer();
		probefactorString.append("probes <- c(");
		for (int x = 1; x <= probelevelSize; x++)
		{
			probefactorString.append("\"P" + x + "\",");
		}
		for (int x = 1; x <= probelevelSize; x++)
		{
			probefactorString.append("\"P" + x + "\",");
		}
		probefactorString.deleteCharAt((probefactorString.length()-1));
		probefactorString.append(")");
		
		System.out.println(probefactorString);
	}

}
