package statistics;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Random;
import java.util.StringTokenizer;

import dataSetDescribers.ExperimentData;
import dataSetDescribers.FilePathGenerator;
import dataSetDescribers.ProbeSetData;


import rScripts.ShellToR;
import utils.ConfigReader;
import utils.StatisticReturnObject;

/*
 * 
 * This is exactly same as the class CyberTPaired.java.  Except that in the method
 * writeCommandFile, we call the method writePairedDataFile and create a ratio of the values (expt/contrl)
 * and enter that into the data file.
 * 
 * CybeT paired requires a ratio in order to run properly.....
 * 
 * Therefore is using the CyberTPaired.java class, make sure your data are ratios. Or just use this.
 * 
 * author = Rob Reid
 */
public class CyberTPairedTakeUnloggedDifference implements StatisticsProvider
{
	private List<Double> tValues =new ArrayList<Double>();
	private List<Double> pValues = new ArrayList<Double>();
	private HashMap<String, Integer> positions = new HashMap<String, Integer>();
	private final boolean flipAtRandom;
	private final Random random = new Random();
	
	public CyberTPairedTakeUnloggedDifference(boolean flipAtRandom)
	{
		this.flipAtRandom = flipAtRandom;
	}
	
	public CyberTPairedTakeUnloggedDifference()
	{
		this(false);
	}
	
	private void parseResultsFile( File resultsFile, ExperimentData experiment ) throws Exception
	{
		BufferedReader reader = new BufferedReader(new FileReader(resultsFile));
		
		reader.readLine();
		
		String nextLine = reader.readLine();
		
		while( nextLine != null && nextLine.trim().length() != 0)
		{
			StringTokenizer sToken = new StringTokenizer(nextLine, ",");
			String indexString = sToken.nextToken();
			indexString = indexString.substring(1);
			indexString = indexString.substring(0, indexString.length() -1);
			
			int index = Integer.parseInt(indexString)-1;
			
			if (tValues.size() != index )
				throw new Exception("Parsing error");
			
			if ( pValues.size() != index)
				throw new Exception("Parsing error");
			
			positions.put( experiment.getProbeSetData().get(index).getProbeSetId(), index );
			
			//int numToSkip = experiment.getNumberOfBaseline() + 6;    //This is needed for Julies and for latin square
			int numToSkip = experiment.getNumberOfBaseline() + 6;
			
			for ( int x=1; x <= numToSkip ; x++)
				sToken.nextToken();
			
			tValues.add( Double.parseDouble(sToken.nextToken()));
			pValues.add( Double.parseDouble(sToken.nextToken()));
			
			nextLine = reader.readLine();
		}
		reader.close();
	}
	
	private List<Double> getSumList(ExperimentData experiment) throws Exception
	{
		if( experiment.getNumberOfBaseline() != experiment.getNumberOfExperiments() )
			throw new Exception("Paired data must have equal number of chips in both conditions");
		
		List<ProbeSetData> probeSetData = experiment.getProbeSetData();
		
		List<Double> sums = new ArrayList<Double>();
		
		for( int x=0; x < experiment.getNumberOfBaseline(); x++ )
		{
			double sum =0 ;
			
			for( ProbeSetData data : probeSetData )
				sum+= Math.pow(2, data.getBaselineData().get(x));
			
			sums.add( sum );
		}
		
		for( int x=0; x < experiment.getNumberOfExperiments(); x++ )
		{
			double sum =0 ;
			
			for( ProbeSetData data : probeSetData )
				sum+= Math.pow(2, data.getExperimentData().get(x));
			
			sums.add( sum );
		}
		
		return sums;
	
	}
	
	private List<Boolean> getShuffledList(int listSize ) 
	{
		List<Boolean> list = new ArrayList<Boolean>();
		
		for( int x=0; x < listSize /2; x++)
		{
			list.add(true);  list.add(false);
		}
		
		if( listSize % 2 == 1 )
			list.add(random.nextBoolean());
		
		Collections.shuffle(list);
		
		return list;
	}
	
	private void writePairedDataFile( File dataFile,  ExperimentData experiment ) throws Exception
	{
		BufferedWriter writer = new BufferedWriter(new FileWriter(dataFile));
		
		List<Double> sums = getSumList(experiment);
		
		List<Boolean> shuffledList = null;
		
		if( flipAtRandom)
			shuffledList = getShuffledList( experiment.getNumberOfBaseline() );
		
		for( ProbeSetData data : experiment.getProbeSetData() )
		{	
			double columnSum = 0;
			
			int columnIndex=0;
			
			for (int i = 0; i < data.getBaselineData().size(); i++)
			{
				// the data is in log(2) space 
				double baseLine = Math.pow(2,data.getBaselineData().get(i));
				double exptvalue = Math.pow(2,data.getExperimentData().get(i));
				//we want ln space
				//baseLine = Math.log(baseLine);
				//exptvalue = Math.log(exptvalue);
				
				double lnVal = exptvalue / baseLine;
				
				if( shuffledList != null && shuffledList.get(i))
					lnVal = 1 / lnVal;
				
				writer.write(  lnVal + "\t" );
				//calc the column sum
				//double columstuff = (Math.pow(2,data.getBaselineData().get(i)) + Math.pow(2,data.getExperimentData().get(i))) / sums.get(columnIndex);
				//columnSum+= (  Math.log( columstuff)  ); 
				//columnIndex++;
			}
			
			
			for( Float f : data.getBaselineData() )
			{
				double unloggedVal = Math.pow(2,f); 
				//writer.write( unloggedVal + "\t" );
				columnSum+= (  Math.log( unloggedVal / sums.get(columnIndex))  ); 
				columnIndex++;
			}
			
			for( Float f : data.getExperimentData() )
			{
				double unloggedVal = Math.pow(2,f); 
				//writer.write( unloggedVal + "\t" );
				columnSum+= (  Math.log( unloggedVal / sums.get(columnIndex))  ); 
				columnIndex++;
			}
			double finalColumnVal = columnSum / ( experiment.getNumberOfBaseline() );
			writer.write( finalColumnVal + "\n" );
		}
		
		writer.flush();  writer.close();
	}
	
	private void writeCommandFile( ExperimentData experiment  ) throws Exception
	{
		File commandFile = FilePathGenerator.deleteAndEnsureDeleted(
				ConfigReader.getAffyAnalysisOutputDir() + File.separator + "rCommand.txt");
		
		BufferedWriter writer = new BufferedWriter(new FileWriter(commandFile));
		
		File cyberTFunctions = new File(ConfigReader.getPINCDirectory() + File.separator + 
				"GCRMA-CYBERT-FDR BM.R");
		
		writer.write( "source(\"" +  FilePathGenerator.getRPath(cyberTFunctions)  + "\")\n" );
		
		File headerlessData = FilePathGenerator.deleteAndEnsureDeleted( ConfigReader.getAffyAnalysisOutputDir()
				 + File.separator + "DataForROkToDelete.txt" );
		
		writePairedDataFile(headerlessData, experiment);
		
		writer.write( "datamatrix<-read.table(\"" + FilePathGenerator.getRPath( headerlessData ) + 
				"\")\n");
		
		writer.write(
		"bayesianResultMatrix2 <- bayesT.pair(datamatrix," + experiment.getNumberOfBaseline() 
		+ ",conf=" + (3*experiment.getNumberOfBaseline()) + ")\n");
		
		File rResults = FilePathGenerator.deleteAndEnsureDeleted( ConfigReader.getAffyAnalysisOutputDir()
				+ File.separator + "rResultsOkToDelete.txt");
		
		//System.out.println(" Yay we are here, write that next line !");
		
		writer.write("write.csv(bayesianResultMatrix2, file=\"" + FilePathGenerator.getRPath(rResults)
				+ "\");\n" );
		
		writer.flush();  writer.close();
		System.out.println("Shelling to R");
		ShellToR.shellToR(commandFile);
		System.out.println("Returned from R");
		parseResultsFile(rResults, experiment);
	}
	
	public void preliminaryCalculations(ExperimentData experiment) throws Exception
	{
		writeCommandFile(experiment);
	}

	public StatisticReturnObject getStats(ProbeSetData probeSet) throws Exception
	{
		int index = positions.get(probeSet.getProbeSetId());
		
		return new StatisticReturnObject( tValues.get(index), pValues.get(index));
	}

	public boolean hasPValue()
	{
		return true;
	}

	public String getDescription()
	{
		return "CyberTPaired";
	}
	
	
	public static void main(String[] args) throws Exception
    {
		
    }
	
}
