package statistics;

import java.io.BufferedOutputStream;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.tigr.midas.engine.Lowess;

import parsers.ScoreFileLineParser;

import utils.StatFunctions;
import utils.TTest;

import dataSetDescribers.FilePathGenerator;
import dataSetDescribers.RatioExperiment;

public class StatisticsLevelNormalization
{
	public static final int WINDOW_SIZE = 1000;
		
	private void writeNewScoreFile(List<ScoreFileLineParser> list, RatioExperiment re, 
			 StatisticsProvider sp ) throws Exception
	{
		 DataOutputStream out = new DataOutputStream( new BufferedOutputStream(
	         	new FileOutputStream( FilePathGenerator.getScoreFile(re,sp))));
		
		 out.writeInt(list.size());
		 
		double byConstant = 0;
		
		for ( int x=0; x <  list.size() ; x++)
		{
			byConstant += (1.0/(x+1));
		}
		
		for( int x=0; x< list.size(); x++)
		{
			ScoreFileLineParser sflp = list.get(x);
			out.writeUTF( sflp.getProbeSetId());
			
			if( re.getSpikeDetails() == null || ! re.getSpikeDetails().getAllSpikedGeneIds().contains( 
					sflp.getProbeSetId() ) )
				out.writeBoolean(false );
			else
				out.writeBoolean(true);
			
			out.writeFloat( sflp.getAverage() );
			out.writeDouble( sflp.getScore()  );
			
			out.writeDouble( sflp.getUnnormalizedFoldChange()  );
			
			out.writeDouble( sflp.getPValue() );
			out.writeDouble( ((x+1.0) / list.size() ) );
			out.writeFloat( sflp.getExpectedScore() );
			out.writeFloat( sflp.getUnnormalizedScore() );
			double bhFdr = (list.size() * sflp.getPValue())/(x+1);
			out.writeDouble( bhFdr  );
			out.writeDouble( (byConstant*bhFdr));
		}
		
		out.flush(); out.close();

		//WriteScores.kickOffRocIfAppropriate(re, sp, FilePathGenerator.getScoreFile(re, sp)  );
	}
	
	public static void normalizeList(List<ScoreFileLineParser> list, boolean setPValues) throws Exception
	{
		float[] xInput = new float[list.size()];
		float[] yInput = new float[list.size()];
		
		for( int x=0; x < list.size(); x++ )
		{
			ScoreFileLineParser sflp = list.get(x);
			xInput[x] = sflp.getAverage();
			yInput[x] = sflp.getScore();
		}
		
		Lowess lowess = new Lowess(xInput,yInput,WINDOW_SIZE);
		
		float[] yEst = lowess.getYEst();
		
		List<Float> scores = new ArrayList<Float>();
		for( int x=0; x < list.size(); x++)
		{
			ScoreFileLineParser aLine = list.get(x);
			aLine.setExpectedScore( yEst[x] );
			aLine.setUnnormalizedScore( yInput[x]  );
			aLine.setScore( yInput[x] - yEst[x] );
			scores.add( aLine.getScore() );
		}
		
		if ( setPValues)
		{
			double mu = TTest.getAverage(scores);
			double sd = TTest.getStDev(scores);
			
			for( ScoreFileLineParser sflp : list )
			{	
				double score = sflp.getScore();
				sflp.setPValue( getPValueFromScore(score,mu,sd) );
			}
			
			Collections.sort( list, new ScoreFileLineParser.SortByPValue() );
		}
	}
	
	
	public static double getPValueFromScore( double score, double mu, double sd )
	{
		double statNorm1 = Math.abs(  score  - mu) / sd;
		return 2* StatFunctions.pnorm(statNorm1, true);
	
	}
	
	public static void main(String[] args) throws Exception
	{
		double p1 = StatFunctions.pnorm(.34,true);
		double p2 = StatFunctions.pnorm(.34,false);
		
		System.out.println(p1 + " " + p2 + " " + (p1+p2));
		
		p1 = StatFunctions.pnorm(-.34,true);
		p2 = StatFunctions.pnorm(-.34,false);
		
		System.out.println(p1 + " " + p2 + " " + (p1+p2));
		
	}
	
	public StatisticsLevelNormalization( RatioExperiment re, StatisticsProvider sp   ) 
	throws Exception
	{
		if( ! re.isUseStatisticLevelNormalization() )
			throw new Exception("This method should only be called when useStatsLevelNormalization is set to true");
		
		re.setUseStatisticLevelNormalization(false);
		
		File unnormalizedScoreFile = FilePathGenerator.getScoreFile(re,sp);
		
		if( ! unnormalizedScoreFile.exists() )
			throw new Exception("Prelimary file missing " + 
					unnormalizedScoreFile.getAbsolutePath());
		
		List<ScoreFileLineParser> list = ScoreFileLineParser.parseFile(unnormalizedScoreFile);
		Collections.sort(list, new ScoreFileLineParser.SortByAverage());
			
		normalizeList(list, true);
		re.setUseStatisticLevelNormalization(true);
		writeNewScoreFile(list, re, sp);
		
	}
}
